function [Result,Data] = AnalyzeModels(Samples)

%SETTINGS
% Compares growth to WT
Threshold = 0.3; %Growth rate fraction below Threshold is seen as essential.

%find base path
location = '../S4-Models/';

Data = {
    ...File paths   ,       {Model names                                       , TestData
    {location, {'Input/Pputida/iJN746_COMMGEN','Input/Pputida/iJP962_COMMGEN','BCM/Pputida/BCM_iJN746_COMMGEN','BCM/Pputida/BCM_iJP962_COMMGEN','ACM/Pputida/ACM_iJN746_COMMGEN','ACM/Pputida/ACM_iJP962_COMMGEN'},'PputidaTests'} ... Pseudomonas putida
    {location, {'Input/Bsubtilis/iBsu1103_COMMGEN','Input/Bsubtilis/iYO844_COMMGEN','BCM/Bsubtilis/BCM_iBsu1103_COMMGEN','BCM/Bsubtilis/BCM_iYO844_COMMGEN','ACM/Bsubtilis/ACM_iBsu1103_COMMGEN','ACM/Bsubtilis/ACM_iYO844_COMMGEN'},'BsubtilisTests'} ... Bacillus subtilis
    {location, {'Input/Mtuberculosis/iNJ661_COMMGEN','BCM/Mtuberculosis/BCM_iNJ661_COMMGEN','ACM/Mtuberculosis/ACM_iNJ661_COMMGEN'},'MtuberculosisTests'} ... Mycobacterium tuberculosis
    {location, {'Input/Scerevisiae/iND750_COMMGEN','Input/Scerevisiae/iMM904_COMMGEN','BCM/Scerevisiae/BCM_iND750_COMMGEN','BCM/Scerevisiae/BCM_iMM904_COMMGEN','ACM/Scerevisiae/ACM_iND750_COMMGEN','ACM/Scerevisiae/ACM_iMM904_COMMGEN'},'ScerevisiaeTests'} ... Saccharomyces cerevisiae    
    };

mediumFile = 'GrowthExperiments.xlsx';

%Process inputs
if nargin<1
    Samples = 1:numel(Data);
end

% Set variables
changeCobraSolver('gurobi5')
Result = struct;

for k=Samples %For each organism
    % Load data
    %General information
    OrgData = Data{k};
    Path = OrgData{1};
    Models = OrgData{2};
    DataSheet = OrgData{3};
    
    %Paths to models
    ModelPaths = strcat(Path,Models);
    
    %Experimental data
    [~, ~, raw] = xlsread(mediumFile,DataSheet);
    raw(1,:) = [];
    strings = raw(:,[1:6 9]);
    numbers = raw(:,7:8);
    strings(cellfun(@isnumeric,strings)) = {''}; %Empty cells are read as [NaN]
    
    Medium   = strings(:,2); % Medium composition used in in vitro experiment
    GeneIDs  = strings(:,3); % Name of the knocked out gene
    Rescue   = strings(:,4); % Species added to the medium to rescue growth of the mutant
    MNXRef   = strings(:,5); % ID(s) of the species in the MnXRef namespace
    InVitroGrowth   = strings(:,6); % Whether the bacteria grow
    ParRes = strings(:,7);  % Indicates expected result of parent experiment -> if incorrect, no point in performing.
    
    ExpNo  = cell2mat(numbers(:,1));   % Indicates experiment hierarchy 
    Parent = cell2mat(numbers(:,2));   % Indicates experiment hierarchy
    
    %Define variables
    NumTests  = size(raw,1);
    NumModels = numel(Models);
    PredictGrowth  = nan(NumTests,NumModels);
    PredictTrue    = false(NumTests,NumModels);
    PredictFalse   = false(NumTests,NumModels);
    WTGrowth = nan(NumTests,NumModels);
    
    dM = nan(1,NumModels);
    iM = nan(1,NumModels);
    dR = nan(1,NumModels);
    iR = nan(1,NumModels);
    NumR = nan(1,NumModels);
    NumM = nan(1,NumModels);
    
    for j=1:NumModels % For each model
        % Obtain growth predictions
        %Load model
        model = loadMatModel(ModelPaths{j});
        model = ResolveNaNs(model);
        %obtain number of dead end and blocked reactions / metabolites
        MnXAnalysis = MnXModelStructuralAnalysis;
        MnXAnalysis.MnXModelDeadEnds(model,true);
        MnXAnalysis.MnXModelDeadEnds(model,false);
        
        dM(j) = count(MnXAnalysis.deadEnds.iM);
        iM(j) = count(MnXAnalysis.deadEndsRec.iM);
        dR(j) = count(MnXAnalysis.deadEnds.iR);
        iR(j) = count(MnXAnalysis.deadEndsRec.iR);
        NumR(j) = model.stoich.r;
        NumM(j) = model.stoich.m;
        
        %Convert to cobra model
        cobra = ModelToCobra(model);
        UpGeneList = upper(cobra.genes); %List of genes in the model, all caps
        SpecInModel = regexprep(cobra.mets,'\[.*',''); %List of metabolites in the model, decompartmentalised
        
        %Predict growth
        TestGrowth(cobra);
        
        %Save results if last sample
        if j==NumModels
            SaveResult;
        end
    end
end

    function TestGrowth(cobra)
        %Obtain model growth predictions and compare to experimental data
        
        for i=1:(numel(GeneIDs))
            i
            Continue = true; %Reset
            
            %Set medium composition
            if i==1 || ~strcmp(Medium(i),Medium(i-1)); %If either no medium has been set yet OR a different medium is to be used.
                cobra = SwitchMedium(cobra,Medium{i},mediumFile);
                WTBase = cobra;
            end
            
            %Check if parent experiment, if any, was succesful
            if Parent(i)~=0
                Par = ExpNo==Parent(i);
                if PredictGrowth(Par,j) <= 10^-6 
                    ParPred = 'No';
                elseif PredictGrowth(Par,j) > 10^-6
                    ParPred = 'Yes';
                else %nan
                    ParPred = 'NAN';
                end
                CorrectPred = ParRes(i);
                if ~strcmp(ParPred,CorrectPred)
                    Continue = false;
                end
            end
            
            %Adjust model to mutant
            if Continue && ~strcmp(GeneIDs(i),'WT') % if mutant strain
                UpMutGenes = unique(upper(strsplit(GeneIDs{i},' ')));
                GeneInModel = any(ismember(UpMutGenes,UpGeneList));
                
                %Test whether gene is in model
                if GeneInModel
                    KOGenes = cobra.genes(ismember(UpGeneList,UpMutGenes)); %Find gene names in model
                    tmpmodel = deleteModelGenes(cobra,KOGenes); %disable reactions which require the knocked-out gene
                else
                    Continue = false;
                end
                
                %Track that we're dealing with a mutant strain
                Mutant = true;
            elseif Continue %WT strain
                tmpmodel = cobra;
                Mutant = false;
            end
            
            %Check if any compounds should be added to the medium
            if Continue
                CpdAdded   = ~strcmp('NA',Rescue(i)); % NA is used as value if nothing is added.
                if CpdAdded
                    AddCpds = splitstr(MNXRef{i},';');
                    CpdInModel = any(ismember(AddCpds,SpecInModel));
                    if CpdInModel
                        tmpmodel = addMediumComponent(tmpmodel,strcat(AddCpds,'[MNXC2]')); %MNXC2 corresponds to the extracellular environment
                        WTmodel =  addMediumComponent(WTBase,strcat(AddCpds,'[MNXC2]')); %MNXC2 corresponds to the extracellular environment
                    else
                        Continue = false;
                    end
                else
                    WTmodel = WTBase;
                end
            end
            
            if Continue
                sol = optimizeCbModel(tmpmodel);
                PredictGrowth(i,j) = sol.f;
                %In case we're dealing with a mutant strain, also test WT
                if Mutant
                    WTsol = optimizeCbModel(WTmodel);
                    WTGrowth(i,j) = WTsol.f;
                else
                    WTGrowth(i,j) = sol.f; %Identical to WT
                end
                
                %Convert growth predictions to boolean predictions
                %depending on experiment type
                if Mutant
                   %Compare to WT
                   if WTGrowth(i,j)<10^-6
                       %No prediction can be made
                       PredictGrowth(i,j) = NaN;
                       WTGrowth(i,j) = NaN;
                   elseif PredictGrowth(i,j)/WTGrowth(i,j)>Threshold
                       PredictTrue(i,j) = true;
                   else
                       PredictFalse(i,j) = true;
                   end
                else
                    %Compare to absolute growth rate
                    if PredictGrowth(i,j)>10^-6
                        PredictTrue(i,j)=true;
                    else
                        PredictFalse(i,j)=true;
                    end
                end
            else
                PredictGrowth(i,j) = NaN;
                WTGrowth(i,j) = NaN;
            end
        end
    end

    function SaveResult
        % Compares predictions to experimental data and writes an overview
        % file
        
        %Experimental result
        True  = strcmp(InVitroGrowth,'Yes');
        False = strcmp(InVitroGrowth,'No');
        
        %Compare for subset of data (predictable by all models)
        idx = ~(any(isnan(PredictGrowth),2));
        TP = sum(PredictTrue(idx,:)  & repmat(True(idx) ,1,NumModels),1);
        FP = sum(PredictTrue(idx,:)  & repmat(False(idx),1,NumModels),1);
        TN = sum(PredictFalse(idx,:) & repmat(False(idx),1,NumModels),1);
        FN = sum(PredictFalse(idx,:) & repmat(True(idx) ,1,NumModels),1);
        % Summary statistics
        Result.(DataSheet).Specificity = TN./(TN+FP);
        Result.(DataSheet).Sensitivity = TP./(TP+FN);
        Result.(DataSheet).Accuracy    = (TP+TN)./(TP+TN+FP+FN);
        Result.(DataSheet).MCC         = (TP.*TN-FP.*FN)./sqrt((TP+FP).*(TP+FN).*(TN+FP).*(TN+FN));
        
        %Compare for all available growth data
        idx(:) = true;
        TP = sum(PredictTrue(idx,:)  & repmat(True(idx) ,1,NumModels),1);
        FP = sum(PredictTrue(idx,:)  & repmat(False(idx),1,NumModels),1);
        TN = sum(PredictFalse(idx,:) & repmat(False(idx),1,NumModels),1);
        FN = sum(PredictFalse(idx,:) & repmat(True(idx) ,1,NumModels),1);
        % Summary statistics
        Result.(DataSheet).ALL_Specificity = TN./(TN+FP);
        Result.(DataSheet).ALL_Sensitivity = TP./(TP+FN);
        Result.(DataSheet).ALL_Accuracy    = (TP+TN)./(TP+TN+FP+FN);
        Result.(DataSheet).ALL_MCC         = (TP.*TN-FP.*FN)./sqrt((TP+FP).*(TP+FN).*(TN+FP).*(TN+FN));
        
        %Create boolean representation
        BoolRep = nan(size(PredictGrowth));
        BoolRep(PredictTrue)  = 1;
        BoolRep(PredictFalse) = 0;
        
        % Create overview
        Rescue = strcat('+',Rescue);
        Rescue(strcmp(Rescue,'+NA')) = {''};
        Output = [strcat(GeneIDs,'_',Rescue),InVitroGrowth,cellfun(@num2str,num2cell(PredictGrowth),'UniformOutput',false)];
        WTData = [strcat(GeneIDs,'_',Rescue),InVitroGrowth,cellfun(@num2str,num2cell(WTGrowth),'UniformOutput',false)]; 
        BoolData = [strcat(GeneIDs,'_',Rescue),InVitroGrowth,cellfun(@num2str,num2cell(BoolRep),'UniformOutput',false)]; 
        
        Output = [[{'GeneKO_AddSpecies','InVitroGrowth'},Models];Output];
        WTData = [[{'GeneKO_AddSpecies','InVitroGrowth'},Models];WTData];
        BoolData = [[{'GeneKO_AddSpecies','InVitroGrowth'},Models];BoolData];
        
        fid = fopen([DataSheet '_KOoverview.tsv'], 'w+');
        for i=1:size(BoolData,1)
            for z=1:size(BoolData,2)-1
                fprintf(fid,'%s\t',BoolData{i,z});
            end
            fprintf(fid,'%s\n',BoolData{i,end});
        end
        fclose(fid);
        
        Result.(DataSheet).dM = dM;
        Result.(DataSheet).iM = iM;
        Result.(DataSheet).dR = dR;
        Result.(DataSheet).iR = iR;
        Result.(DataSheet).NumR = NumR;
        Result.(DataSheet).NumM = NumM;
        
        Result.(DataSheet).Data = Output;
        Result.(DataSheet).WTData = WTData;
        Result.(DataSheet).BoolData = BoolData;
    end
end